// Z88DK DEMO

#include <llvgraphics.h>
#include <graphics.h>
#include <ti/keyboard.h>
#include <stdlib.h>

#ifdef TI82
#define TI "82"
#endif
#ifdef TI83
#define TI "83"
#ifdef __ION__
#define DELAY_OFFSET 7
#else
#error TI83: ION only!
#endif
#endif
#ifdef TI8x
#define TI "83+"
#define DELAY_OFFSET 7
#endif
#ifdef TI85
#define TI "85"
#define DELAY_OFFSET 10
#endif
#ifdef TI86
#define TI "86"
#endif
#define VERSION "1.0"

// #define ABS(x) (((x)>=0)?(x):-(x))
#define ABS abs
#define SGN(x) (((x)>=0)?(x!=0):-1)

// I DO NOT KNOW WHY I NEED THIS, BUT OTHERWISE IT DOES NOT WORK!
#define RAND() (rand() & 0xFF)

typedef int _int;
#define int char

void DELAY (unsigned _int n)
{
 int i;
 while (n--)
  for (i = 0; i != 15; ++i);
}

extern char MCsprites [];

#asm
.smc_MCsprites
    ; SPRITE 0: pointer
	defb	8, 8
	defb	@00011000
	defb	@00011000
	defb	@00011000
	defb	@11100111
	defb	@11100111
	defb	@00011000
	defb	@00011000
	defb	@00011000
    ; SPRITE 1: town
	defb	8, 8
	defb	@01000000
	defb	@01000000
	defb	@01000100
	defb	@11101110
	defb	@11111111
	defb	@11111111
	defb	@11111111
	defb	@11111111
    ; SPRITE 2: station
	defb	8, 8
	defb	@00011000
	defb	@00011000
	defb	@00011000
	defb	@00011000
	defb	@01111110
	defb	@01111110
	defb	@11111111
	defb	@11111111
    ; SPRITE 3: missile
	defb	8, 6
	defb	@00000000
	defb	@00000000
	defb	@00011000
	defb	@00111100
	defb	@00111100
	defb	@00011000
    ; SPRITE 4: explosion
	defb	8, 8
	defb	@00111100
	defb	@01111110
	defb	@11111111
	defb	@11111111
	defb	@11111111
	defb	@11111111
	defb	@01111110
	defb	@00111100
    ; SPRITE 5: black
	defb	8, 8
	defb	@11111111
	defb	@11111111
	defb	@11111111
	defb	@11111111
	defb	@11111111
	defb	@11111111
	defb	@11111111
	defb	@11111111
    ; SPRITE 6: shot
	defb	2, 7
	defb	@11000000
	defb	@11000000
	defb	@11000000
	defb	@11000000
	defb	@11000000
	defb	@11000000
	defb	@11000000
    ; SPRITE 7: pixel
	defb	1, 1
	defb	@10000000
#endasm

#define SPRITE(n) &(MCsprites[n])
#define POINTER SPRITE(0)
#define TOWN SPRITE(10)
#define STATION SPRITE(20)
#define MISSILE SPRITE(30)
#define EXPLOSION SPRITE(38)
#define BLACK SPRITE(48)
#define SHOT SPRITE(58)
#define PIXEL SPRITE(67)

#define BOTTOM 17
#define LastX GetNumX
#define LastY (GetNumY - BOTTOM)
#define StartX (GetMaxX / 2)
#define StartY (GetMaxY - BOTTOM)
#define SHOTS_MAX 20
#define SHOTS_AT_ONCE 3
#define SHOT_NEAR 6

unsigned int TownsLiving [6];
unsigned int NumTowns;
unsigned int NumShots;
unsigned _int Score;

void DrawShot (unsigned int n)
{
 putsprite_later (spr_xor, 1 + 3 * n, LastY + 9, SHOT);
}

char *int2str_score (unsigned _int s)
{
 static char buf[] = "xxxxx0";
 static unsigned int i;
 for (i = 5; i; s /= 10)
  buf [--i] = '0' + s % 10;
 return buf;
}

void DrawScore ()
{
 outtextxy_later (GetMaxX - 30, LastY + 9, int2str_score (Score));
}

void SetScore (unsigned _int n)
{
 DrawScore ();
 Score = n;
 DrawScore ();
}

void InitGame ()
{
 *(unsigned _int *) &TownsLiving [0] = 0x0101;
 *(unsigned _int *) &TownsLiving [2] = 0x0101;
 *(unsigned _int *) &TownsLiving [4] = 0x0101;
 Score = 0;
 NumTowns = 6;
}

void InitScreen ()
{
 unsigned int x;
 cleargraphics ();
 if (TownsLiving [0])
  putsprite_later (spr_xor, GetMaxX * 1 / 8 - 4, LastY, TOWN);
 if (TownsLiving [1])
  putsprite_later (spr_xor, GetMaxX * 2 / 8 - 4, LastY, TOWN);
 if (TownsLiving [2])
  putsprite_later (spr_xor, GetMaxX * 3 / 8 - 4, LastY, TOWN);
 if (TownsLiving [3])
  putsprite_later (spr_xor, GetMaxX * 5 / 8 - 4, LastY, TOWN);
 if (TownsLiving [4])
  putsprite_later (spr_xor, GetMaxX * 6 / 8 - 4, LastY, TOWN);
 if (TownsLiving [5])
  putsprite_later (spr_xor, GetMaxX * 7 / 8 - 4, LastY, TOWN);
 putsprite_later (spr_xor, GetMaxX * 4 / 8 - 4, LastY, STATION);
 for (x = 0; x != (GetNumX / 8); ++x)
 {
  putsprite_later (spr_or, x << 3, LastY + 8, BLACK);
  putsprite_later (spr_or, x << 3, LastY + 9, BLACK);
 } 
}

void RemoveTown (int n)
{
 putsprite_later (spr_xor, (GetMaxX * (n + 1 + (n > 2)) >> 3) - 4, LastY, TOWN);
 TownsLiving [n] = 0;
 --NumTowns;
}

void ClearTown (unsigned int x)
{
 unsigned int n = (x * (unsigned _int) (8) + GetMaxX / 2) / GetMaxX;
 unsigned int x0 = ((GetMaxX * n) >> 3) - 5;    // + one pixel to the left/right
 unsigned int n2;
 if (! (n >= 8 || !n || (unsigned int) (x - x0) > 9))
 {
  if (n == 4)
  {
   n = NumShots;
   while (NumShots && (n - NumShots < 10))
    DrawShot (--NumShots);
  }
  else
  {
   n2 = n - 1 - (n > 4);
   if (TownsLiving [n2])
    RemoveTown (n2);
  }  
 }
}

typedef struct SHOOT
{
 unsigned int x1, y1, x2, y2;
 // x1 == 0 wenn leer
}
SHOOT;

typedef struct ENEMY
{
 unsigned int x0, dy, cy, x, y, exist; // alle mit x!
 signed int dx;
}
ENEMY;

#define MAX_ENEMIES 8

void ClearEnemy (ENEMY *en)
{
 unsigned int y;
 for (y = 1; y < en->y; ++y)
  putsprite_later (spr_xor, en->x0 + (y * en->dx + 2) / 4, y, PIXEL);
 en->exist = 0;
}

unsigned int NumEnemies;
unsigned int i, k;
signed int xdir, ydir;

void DrawSHOOT (struct SHOOT *p)
{
 if (p->x1)
 {
  xdir = p->x1 - p->x2;
  ydir = p->y1 - p->y2;
  putsprite_later (spr_xor, p->x1, p->y1, (ABS(xdir) + ABS(ydir) < SHOT_NEAR) ? EXPLOSION : MISSILE);
 }
}

void MoveSHOOT (struct SHOOT *p, struct ENEMY *Enemies)
{
 if (p->x1)
 {
  xdir = p->x1 - p->x2;
  ydir = p->y1 - p->y2;
  if (ABS(xdir) + ABS(ydir) < SHOT_NEAR)
   for (k = 0; k < MAX_ENEMIES; ++k)
    if (Enemies [k].exist
     && (unsigned int) (Enemies [k].y - p->y1) < 8
     && (unsigned int) (Enemies [k].x - p->x1) < 8)
    {
     putsprite_later (spr_xor, Enemies [k].x, Enemies [k].y, PIXEL);
     ClearEnemy (&Enemies [k]);
     --NumEnemies;
     SetScore (Score + 3);
    } 
  if (!(xdir || ydir))
   p->x1 = 0; 
#if GetMaxX == 95
  if (ABS(xdir) > ABS(ydir))
   p->x1 -= SGN(xdir);
  else
   p->y1 -= SGN(ydir);
#else
  p->x1 -= SGN(xdir);
  p->y1 -= SGN(ydir);           // faster movement on 128px-calcs
#endif
 }
}

void MissileCommand (unsigned int Waves, 
                     unsigned int EnemiesPerWave, 
                     unsigned _int Delay, 
                     unsigned int EnemySpeed)
{
 static unsigned int EndFlag;
 static unsigned int PlayerX, PlayerY;
 static unsigned int ShotState;
 struct SHOOT Shots [SHOTS_AT_ONCE];
 static unsigned int EnemyTime;
 static unsigned int NumWaves;
 struct ENEMY Enemies [MAX_ENEMIES];
 NumShots = SHOTS_MAX;
 NumWaves = Waves;
 
 EndFlag = 0;
 PlayerX = (LastX / 2) & 0xFE;
 PlayerY = (LastY / 2) & 0xFE;

 swapgfxbk ();
 InitScreen ();
 DrawScore ();

 for (i = 0; i < NumShots; ++i)
 {
  DrawShot (i);
 } 
 
 for (i = 0; i < SHOTS_AT_ONCE; ++i)
 {
  Shots[i].x1 = 0;
 }

 for (i = 0; i < MAX_ENEMIES; ++i)
 {
  Enemies[i].exist = 0;
 }
 NumEnemies = 0;
 EnemyTime = 0;

 while (!EndFlag)
 {
  // SHOW
  putsprite_later (spr_xor, PlayerX, PlayerY, POINTER);

  for (i = 0; i != SHOTS_AT_ONCE; ++i)
   DrawSHOOT (&Shots [i]);
  
  swapgfxbk1 ();
  
  DELAY (Delay);

  swapgfxbk ();
  // REMOVE
  putsprite_later (spr_xor, PlayerX, PlayerY, POINTER);
  
  for (i = 0; i != SHOTS_AT_ONCE; ++i)
  {
   DrawSHOOT (&Shots [i]);
   MoveSHOOT (&Shots [i], Enemies);
  } 
  
  if (!NumEnemies)
  {
   if (!(Waves--))
    break;
   for (i = 0; i < EnemiesPerWave; ++i)
   {
    Enemies [i].x0 = (RAND () % (GetNumX - 20)) + 10;
    Enemies [i].dx = (signed int) (RAND () % 7) - 3;
    Enemies [i].dy = RAND() % (4 + (EnemySpeed == 1)) + 3 + (EnemySpeed == 1);
    Enemies [i].exist = 1;
    Enemies[i].cy = Enemies [i].y = 0;
   }
   NumEnemies = EnemiesPerWave;
   for (i = NumShots; i + 13 < NumShots + EnemiesPerWave + (NumTowns << 1)
                   && i < SHOTS_MAX; ++i)
    DrawShot (i);
   NumShots = i;
  }
  
  if (!(++EnemyTime % EnemySpeed))
  {
   EnemyTime = 0;
   for (i = 0; i < EnemiesPerWave; ++i)
   {
    struct ENEMY *e = &Enemies [i];
    if (e->exist)
    {
     ++e->cy;
     if (!(e->cy %= e->dy))
     {
      ++e->y;
      e->x = e->x0 + (e->y * e->dx + 2) / 4;
      if (e->x <= GetMaxX)
      {
       if (e->y > LastY + 4)
       {
        ClearTown (e->x);
        ClearEnemy (e);
        --NumEnemies;
       }
       else
        putsprite_later (spr_xor, e->x, e->y, PIXEL);
      }
      else
      {
       ClearEnemy (e);
       --NumEnemies;
      } 
     }
    } 
   }
  } 
  
  // MOVE?
  i = 2;
  do
  {
   if (PlayerX && TI_keypressed (KEY_LEFT))
    --PlayerX;
   if (PlayerY && TI_keypressed (KEY_UP))
    --PlayerY;
   if (PlayerX != ((LastX - 8) & 0xFE) && TI_keypressed (KEY_RIGHT))
    ++PlayerX;
   if (PlayerY != ((LastY - 8) & 0xFE) && TI_keypressed (KEY_DOWN))
    ++PlayerY;
  }
  while (--i && TI_keypressed (KEY_ALPHA));
   
  // SHOOT?
  if (NumShots && (i = TI_keypressed (KEY_2ND)) && !ShotState)
  {
   // i muss true bleiben, sonst nix!
   for (i = SHOTS_AT_ONCE - 1; i != 255; --i)
    if (!Shots[i].x1)
    {
     Shots[i].x1 = StartX - 4;
     Shots[i].y1 = StartY - 4;
     Shots[i].x2 = PlayerX;
     Shots[i].y2 = PlayerY;
     DrawShot (--NumShots);
     break;
    }
   i = (i != 255);              // konnte nicht schiessen => nochmal
  } 
  ShotState = i;
  
  // EXIT?
  EndFlag = TI_keypressed (KEY_QUIT);
 }
 
 while (NumShots)
 {
  DrawShot (--NumShots);
  SetScore (Score + 1);
  swapgfxbk1 ();
  swapgfxbk ();
 }
 
 for (i = 0; i < 6; ++i)
  if (TownsLiving [i])
  {
   RemoveTown (i);
   ++NumTowns;
   TownsLiving [i] = 1;
   SetScore (Score + EnemiesPerWave * (unsigned _int) NumWaves * 6 / EnemySpeed);
   swapgfxbk1 ();
   swapgfxbk ();
   DELAY (250);
  }
  
 outtextxy_later (1, GetNumY - 8, "Press ENTER...");
  
 swapgfxbk1 ();
 
 while (!TI_keypressed (KEY_ENTER));
 while (TI_keypressed (KEY_ENTER));
 
 swapgfxbk ();
 cleargraphics ();
 swapgfxbk1 ();

 if (EndFlag)
  NumTowns = 0; 
}

#define OFFSET ((GetNumX - 96) / 5)

static unsigned int Levels [] =
{
 2, 2,
 4, 2,
 2, 3,
 5, 3,
 2, 4,
 6, 4,
 2, 5,
 6, 5,
 2, 6,
 7, 6,
 2, 7,
 7, 7,
 2, 8,
 8, 8,
 0, 0
};

void main ()
{
 static unsigned int i;
 static unsigned _int randcnt;
 static char Speed[] = "2";
 static unsigned int lvl;
 static unsigned _int Record = 0;
 outtextxy_setfont4;
 outtextxy_setmode (spr_xor);
 for (;;)
 { 
  swapgfxbk ();
  cleargraphics ();
  outtextxy_later (1, 1,  "Missile Command " TI " v" VERSION);
  outtextxy_later (1, 9,  "by Rudolf Polzer");
  outtextxy_later (1, 17, "remake of an ATARI game");
  outtextxy_later (1, 33, "written in SmallC+");
  outtextxy_later (1, 43, "High score:");
  outtextxy_later (GetMaxX - 30, 43, int2str_score (Record));
  for (;;)
  {
   ++randcnt;
   for (i = 0; i < GetNumX; i += 8)
    putsprite_later (spr_or, i, GetMaxY - 8, BLACK); 
   for (i = 0; i < GetNumX; i += 8)
    putsprite_later (spr_or, i, GetMaxY - 7, BLACK); 
   i = 0;
   outtextxy_later (GetMaxX * 0 / 5 + 8 + OFFSET, GetMaxY - 7, "-");
   outtextxy_later (GetMaxX * 1 / 5 + 8 + OFFSET, GetMaxY - 7, Speed);
   outtextxy_later (GetMaxX * 2 / 5 + 8 + OFFSET, GetMaxY - 7, "+");
   outtextxy_later (GetMaxX * 3 / 5 + 4 + OFFSET, GetMaxY - 7, "Go!");
   outtextxy_later (GetMaxX * 4 / 5 + 2 + OFFSET, GetMaxY - 7, "Quit");
   swapgfxbk1 ();
   BEGINLOOP:
    if (TI_keypressed (KEY_F1))
    {
     while (TI_keypressed (KEY_F1));
     if (*Speed != '1')
      --*Speed;
     goto ENDLOOP; 
    }
    if (TI_keypressed (KEY_F3))
    {
     while (TI_keypressed (KEY_F3));
     if (*Speed != '3')
      ++*Speed;
      goto ENDLOOP;
    }
    if (TI_keypressed (KEY_F4))
    {
     while (TI_keypressed (KEY_F4));
     break;
    }
    if (TI_keypressed (KEY_F5))
    {
     swapgfxbk1 ();
     return;
    } 
   goto BEGINLOOP;      // this is NOT a real loop to make break work!
   ENDLOOP:
  } 
  srand (randcnt); 
  InitGame ();
  lvl = 0;
  while (NumTowns && Levels [lvl << 1])
  {
   MissileCommand (Levels [lvl << 1], Levels [(lvl << 1) + 1], DELAY_OFFSET, '4' - *Speed);
   ++lvl;
  }
  if (Record < Score)
   Record = Score;
 } 
}
